
#include <avr/io.h>
#include <avr/interrupt.h>

#include "system.h"

#include "_enc28j60.h"
#include "enc28j60.h"
#include "ip.h"

#include "lcd.h"

#include <stdlib.h>
#include <string.h>

#define THERMOMETER 0

//

#define TIMEZONE (9 * 60 * 60)
#define NTP_SERVER "ntp.nict.jp"

//

#if 0
#define STATIC_IP_ADDRESS 192,168,0,30
#define SUBNET_MASK 255,255,255,0
#define GATEWAY_ADDR 192,168,0,1
#define DNS_SERVER 192,168,0,1
#endif

static const unsigned char macaddr[] = {
	0xfe, 0xff, 0xff, 0x00, 0x00, 0x01
};

//

void print_number(unsigned int v, int n)
{
	int i;
	char tmp[20];
	char *p = tmp + sizeof(tmp);
	*--p = 0;
	for (i = 0; i < n; i++) {
		*--p = v % 10 + '0';
		v /= 10;
	}
	lcd_print(p);
}

void lcd_print_xnumber(unsigned long v, int n)
{
	int i;
	char tmp[10];
	char *p = tmp + sizeof(tmp) - 1;
	*p = 0;
	for (i = 0; i < n; i++) {
		*--p = "0123456789ABCDEF"[v & 15];
		v /= 16;
	}
	lcd_print(p);
}

//

#if 1

void led_ck(int f)
{
	if (f) {
		PORTD |= 0x01;
	} else {
		PORTD &= ~0x01;
	}
}

void led_do(int f)
{
	if (f) {
		PORTD |= 0x02;
	} else {
		PORTD &= ~0x02;
	}
}


void led_send(int v)
{
	int i;
	for (i = 0; i < 8; i++) {
		led_ck(0);
		if (v & 0x80) {
			led_do(1);
		} else {
			led_do(0);
		}
		wait_us(5);
		v <<= 1;
		led_ck(1);
		wait_us(5);
	}
	led_do(1);
}

void led_display(int num, int val)
{
	led_send((num << 4) | val);
}

#endif

#if THERMOMETER

void adc_cs(int f)
{
	if (f) {
		PORTD |= 0x04;
	} else {
		PORTD &= ~0x04;
	}
}

void adc_ck(int f)
{
	if (f) {
		PORTD |= 0x08;
	} else {
		PORTD &= ~0x08;
	}
}

void adc_do(int f)
{
	if (f) {
		PORTD |= 0x10;
	} else {
		PORTD &= ~0x10;
	}
}

int adc_di()
{
	return (PIND & 0x20) ? 1 : 0;
}

unsigned short adc_transfer(unsigned short v)
{
	int i;
	for (i = 0; i < 16; i++) {
		adc_ck(0);
		if (v & 0x8000) {
			adc_do(1);
		} else {
			adc_do(0);
		}
		wait_us(2);
		v <<= 1;
		if (adc_di()) {
			v |= 1;
		}
		adc_ck(1);
		wait_us(2);
	}
	adc_do(1);
	return v;
}

unsigned short sample_ch0()
{
	adc_cs(0);
	int v = adc_transfer(0x6800) & 0x03ff;
	adc_cs(1);
	return v;
}

unsigned short sample_ch1()
{
	adc_cs(0);
	int v = adc_transfer(0x7800) & 0x03ff;
	adc_cs(1);
	return v;
}

#endif

//

void enc28j60_cs(int f)
{
	if (f) {
		PORTB |= 0x04;
	} else {
		PORTB &= ~0x04;
	}
}

int enc28j60_io(int v)
{
#if 01
	SPDR = v;
	while (!(SPSR & 0x80));
	return SPDR;
#else
	return spi_transfer(v);
#endif
}

//

void eth_init(unsigned char const *macaddr)
{
	enc28j60_init(macaddr);
}

unsigned int eth_recv_packet(void *ptr, int maxlen)
{
	int len = enc28j60_peek_packet();
	if (len == 0) {
		return 0;
	}
//	if (len < 0 || len > maxlen) { // 2013-09-07
	if (len < 0) {
		enc28j60_drop_packet();
		return 0;
	}
	enc28j60_recv_packet((unsigned char *)ptr, maxlen);
	return len;
}

void eth_send_packet(void const *ptr, unsigned int len)
{
	enc28j60_send_packet((unsigned char const *)ptr, len);
}

//

#define NTP_PACKET_SIZE 48

void send_ntp_request(unsigned char const *ntp_server_addr)
{
#if 0
	static const unsigned char data[] = {
		0xdb, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	};
#else
	unsigned char data[NTP_PACKET_SIZE];
	memset(data, 0, sizeof(data));
	data[0] = 0xdb;
#endif
	send_udp_packet(ntp_server_addr, 123, 1024, data, sizeof(data));
}

//

bool parse_ntp_packet(struct packet_header_t *packet, unsigned long *p_s, int *p_ms)
{
	if (!packet) {
		return false;
	}
	if (packet->src_port != 123 && packet->dst_port != 123) {
		return false;
	}
	if (packet->length < NTP_PACKET_SIZE) {
		return false;
	}
	unsigned long s = (((unsigned long)packet->data[0x28] << 24) | ((unsigned long)packet->data[0x29] << 16) | ((unsigned long)packet->data[0x2a] << 8) | (unsigned long)packet->data[0x2b]);
	unsigned long ms = (((unsigned long)packet->data[0x2c] << 24) | ((unsigned long)packet->data[0x2d] << 16) | ((unsigned long)packet->data[0x2e] << 8) | (unsigned long)packet->data[0x2f]);
	ms >>= 16;
	ms = (ms * 1000 + 32768) >> 16;
	*p_s = s;
	*p_ms = (int)ms;
	return true;
}

//

int temperature;

#if THERMOMETER

bool process_thermometer_packet(struct packet_header_t *packet, int tempr)
{
	if (!packet) {
		return false;
	}
	if (packet->dst_port != 30) {
		return false;
	}
	char data[2];
	data[0] = tempr >> 8;
	data[1] = tempr & 0xff;
	send_udp_packet(packet->src_addr, packet->src_port, 1024, data, sizeof(data));
}

int get_temperature()
{
	return ((sample_ch0() * 50000L - (424 * 40 * 1023L)) * 10 / (25 * 1023) + 5) / 10;
}

#endif

//

#if 0
unsigned long convert_ymd_to_cjd(int year, int month, int day)
{
	if (month < 3) {
		month += 9;
		year--;
	} else {
		month -= 3;
	}
	year += 4800;
	int c = year / 100;
	return c * 146097 / 4 + (year - c * 100) * 1461 / 4 + (153 * month + 2) / 5 + day - 32045;
}
#endif

void convert_cjd_to_ymd(unsigned long j, int *year, int *month, int *day)
{
	int y, m, d;
	y = (j * 4 + 128179) / 146097;
	d = (j * 4 - y * 146097 + 128179) / 4 * 4 + 3;
	j = d / 1461;
	d = (d - j * 1461) / 4 * 5 + 2;
	m = d / 153;
	d = (d - m * 153) / 5 + 1;
	y = (y - 48) * 100 + j;
	if (m < 10) {
		m += 3;
	} else {
		m -= 9;
		y++;
	}
	*year = y;
	*month = m;
	*day = d;
}

//

void display_date_time(int year, int month, int day, int hour, int minute, int second, int weekday, int temper)
{
	led_display(0, second % 10);
	led_display(1, second / 10 % 10);
	led_display(2, minute % 10);
	led_display(3, minute / 10 % 10);
	led_display(4, hour % 10);
	led_display(5, hour / 10 % 10);
	led_display(6, day % 10);
	led_display(7, day / 10 % 10);
	led_display(8, month % 10);
	led_display(9, month / 10 % 10);
	led_display(10, year % 10);
	led_display(11, year / 10 % 10);
	led_display(12, year / 100 % 10);
	led_display(13, year / 1000 % 10);

	lcd_cls();
	print_number(year, 4);
	lcd_putc('-');
	print_number(month, 2);
	lcd_putc('-');
	print_number(day, 2);
	lcd_newline();
	print_number(hour, 2);
	lcd_putc(':');
	print_number(minute, 2);
	lcd_putc(':');
	print_number(second, 2);

	lcd_locate(11, 0);
	lcd_putc('(');
	switch (weekday) {
	case 0:
		lcd_putc('S');
		lcd_putc('U');
		lcd_putc('N');
		break;
	case 1:
		lcd_putc('M');
		lcd_putc('O');
		lcd_putc('N');
		break;
	case 2:
		lcd_putc('T');
		lcd_putc('U');
		lcd_putc('E');
		break;
	case 3:
		lcd_putc('W');
		lcd_putc('E');
		lcd_putc('D');
		break;
	case 4:
		lcd_putc('T');
		lcd_putc('H');
		lcd_putc('U');
		break;
	case 5:
		lcd_putc('F');
		lcd_putc('R');
		lcd_putc('I');
		break;
	case 6:
		lcd_putc('S');
		lcd_putc('A');
		lcd_putc('T');
		break;
	}
	lcd_putc(')');

#if THERMOMETER
	if (temper < 0) {
		temper = 0;
	}
	if (temper > 999) {
		temper = 999;
	}
	lcd_locate(10, 1);
	int c = temper / 100 % 10;
	if (c == 0) {
		c = ' ';
	} else {
		c += '0';
	}
	lcd_putc(c);
	lcd_putc(temper / 10 % 10 + '0');
	lcd_putc('.');
	lcd_putc(temper % 10 + '0');
	lcd_putc(0xdf);
	lcd_putc('C');
#endif
}

//

void set_led(int f)
{
	if (f) {
		PORTB = 0x01;
	} else {
		PORTB = 0x00;
	}
}

void giveup()
{
	while (1) {
		set_led(1);
		sleep(50);
		set_led(0);
		sleep(50);
	}
}

int main()
{
	unsigned char ntp_server_addr[4];
	unsigned long now;
	bool valid_time;
	bool adjust;
	temperature = 0;

	system_init();

	PORTB = 0x00;
	PORTD = 0x03;
	DDRB = 0x2d;
	DDRC = 0x3f;
	DDRD = 0x1f;

	SPCR = 0x50;
	SPSR = 0x01;

	wait_ms(10);
	sei();

	lcd_init(16, 2);

#if 0
	while (1) {
		set_led(1);
		sleep(50);
		set_led(0);
		sleep(50);
	}
#endif

#if 0
	enc28j60_cs(0);
	spi_transfer(0);
	lcd_print_xnumber(spi_transfer(0), 2);
	enc28j60_cs(1);
	while (1);
#endif

#if THERMOMETER
	static int temparray[16];
	static int tempindex = 0;
	{
		int i;
		for (i = 0; i < 16; i++) {
			temparray[i] = get_temperature();
		}
	}
#endif

	lcd_println("NTP Clock");
	lcd_println("Starting...");

	ip_stack_init(macaddr);
	
#ifdef STATIC_IP_ADDRESS
	{
		unsigned char ip_address[] = {
			STATIC_IP_ADDRESS
		};
		unsigned char subnet_mask[] = {
			SUBNET_MASK
		};
		unsigned char gateway_addr[] = {
			GATEWAY_ADDR
		};
		unsigned char dns_server[] = {
			DNS_SERVER
		};
		ip_config(ip_address, subnet_mask, gateway_addr, dns_server);
	}
#else
	if (!ip_config_with_dhcp()) {
		giveup();
	}
#endif

	if (!gethostbyname(NTP_SERVER, ntp_server_addr)) {
		giveup();
	}

	now = 0;
	valid_time = false;
	adjust = true;

	while (1) {
		if (adjust) {
			set_led(0);
			send_ntp_request(ntp_server_addr);
			adjust = false;
		}

		ip_stack_process();

		struct packet_header_t *packet = take_udp_packet();
		if (packet) {
			unsigned long s;
			int ms;
			if (parse_ntp_packet(packet, &s, &ms)) {
				set_time(s, ms);
				valid_time = true;
				set_led(1);
			}
#if THERMOMETER
			else if (process_thermometer_packet(packet, temperature)) {
				//
			}
#endif
			free(packet);
		}

		if (valid_time) {
			unsigned long t = get_time();
			if (now < t) {
				now = t;
				t += TIMEZONE;
				unsigned long cjd = t / 24 / 60 / 60 + 2415021;
				int hour = t / 3600 % 24;
				int minute = t / 60 % 60;
				int second = t % 60;
				int year;
				int month;
				int day;
				convert_cjd_to_ymd(cjd, &year, &month, &day);

				int temper = 0;
#if THERMOMETER
				temperature = 0;
				int i;
				for (i = 0; i < 16; i++) {
					temperature += temparray[i];
				}
				temperature = (temperature + 8) / 16;
#endif
				display_date_time(year, month, day, hour, minute, second, (cjd + 1) % 7, temperature);
				if (minute % 30 == 29 && second == 30) {
					adjust = true;
				}

#if THERMOMETER
				temparray[tempindex] = get_temperature();
				tempindex = (tempindex + 1) % 16;
#endif
			}
		}

	}

	return 0;
}

